/**
 * Copyright (c) 2006-2008 iWESUN (ShenZhen) Inf.
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * This file is part of the AvrcX MTOS
 * 
 * Author: Winter Hu  <winter.hu@gmail.com>
 * Create: Jan 3, 2007
 */

#include "spi.h"
#include "avrcx.h"

#if __SPI__ == 1

MUTEX(spi_mutex);
EVENT(spi_complete);
SPICB spicb = {NULL, NULL, 0,0, &spi_complete};

/**
 * Initialize SPI for Rx and Tx
 * You should invoke this method to initializ SPI
 *
 */
void init_spi(void){

  #if SPI_MASTER == 1
  unsigned char tmp = (_BV(SPI_SCK) | _BV(SPI_MOSI) | _BV(SPI_SS));
  SPI_DDR  |= tmp;
  SPI_PORT |= (tmp | _BV(SPI_MISO));
  #endif
  
  SPCR = SPI_SPCR;
  #if SPI_SPEED2 == 1
  SPSR = _BV(SPI2X);
  #endif

}

/**
 * Start SPI to send and recv data
 * This method will blocked until all data are sent
 *
 * @param unsigned char* txBuf, The pointer to the send buffer
 * @param unsigned char* rxBuf, The pointer to the recv buffer
 * @param unsigned char size, buffer length
 */
void spi_write(unsigned char* txBuf, unsigned char* rxBuf, 
		     unsigned char size){
  wait_object(&spi_mutex);
  spi_setbuf(txBuf, rxBuf, size);
  spi_start();
  release_mutex(&spi_mutex);
}

/**
 * Fixed SPI Tx and Rx buffer
 *
 * @param unsigned char* txBuf, The pointer to the send buffer
 * @param unsigned char* rxBuf, The pointer to the recv buffer
 * @param unsigned char size, buffer length
 */
void spi_setbuf(unsigned char* txBuf, unsigned char* rxBuf,
			  unsigned char size){
  spicb.txbuf = txBuf;
  spicb.rxbuf = rxBuf;
  spicb.count = size;
}

/**
 * Start SPI
 * This method will blocked until all data are sent
 */
void spi_start(void){

  into_critical();

  cbi(SPI_PORT, SPI_SS);
  
  /* Not following SPI standard*/
  spicb.point = 0;
  sbi(SPI_PORT, SPI_SS);

  exit_critical();
 
  SPDR = *(spicb.txbuf);
  
  wait_object(&spi_complete);
  reset_event(&spi_complete);

  // sbi(SPI_PORT, SPI_SS); // SPI standard
}

/**
 * Serial Transfer Complete
 */
void NAKED SIG_SPI(void);

void SIG_SPI(void){
  into_critical();

  prologue();

  sbi(SPCR, MSTR); // Avoid master be changed to slave by somehow

  // Read
  *(spicb.rxbuf + spicb.point) = SPDR;
  
  spicb.point++;
  if (spicb.point < spicb.count){
    // Write
    SPDR = *(spicb.txbuf + spicb.point);
  }else{
    // Complete
    spicb.point = 0;
    trigger_event(&spi_complete);
  }

  epilogue();
}

#endif /* __SPI__ = 1*/
